//************************************************************
//  University of Patras 
//	Department of Electrical Engineering & Computer Technology
//	 
//	Course: Parallel Programming in AI 
//	Week 03: Collatz conjecture
//
//  Description: How many steps is required starting from an arbitrary
//                  natural number to reach 1?
//        The collatz algorithm: 
//          Start from the number n
//          if n is even then n <- n/2 else n <- 3n+1
//          if n equal 1 stop else repeat previous if            
//
//      example: 12, 6, 3, 10, 5, 16, 8, 4, 2, 1.
//
//  	  collatz03.c: estimate the most frequent number of steps reaching 1 
//                      using the collatz rule. Better pthreads
//               Vector optimization & pthreads ....  
//	Author: Evangelos Dermatas
//	v0.1
/*
*********************************************************************
gcc -v:	gcc version 7.4.0 (Ubuntu 7.4.0-1ubuntu1~18.04.1)
OS: 	Linux Ubuntu 18.04
CPU:	8-core, Intel(R) Core(TM) i7-2630QM CPU @ 2.00GHz, 3991 BogoMips
*********************************************************************
Compiling:	gcc collatz03.c -o collatz03 -fopt-info-vec-optimized
Executing:	time ./collatz03

Output in my Laptop:

Too long to wait ....


*/

// *******************************************************************
#pragma GCC optimize("O3","unroll-loops","omit-frame-pointer","inline") 
#pragma GCC option("arch=native","tune=native","no-zero-upper") 
//************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

// Maximum number of steps
#define NSTEPS  100000
// Maximum natural number

#define N       100000000
unsigned long int Hist[8][NSTEPS+1] ;

// ***************************************
void *CollatzSerial(void *s)
{
    int j = *((int *) s) ;
    int q, k ;

    for( int i = j+1 ; i < N ; i+=8  ) {
        for ( k = 0, q = i ; k < NSTEPS ; k++ ) {
            if ( q == 1 ) {
                Hist[j][k]++ ;
                break ;
            }
            if ( q%2 == 0 )
                q /= 2 ;
            else
                q=3*q+1;
        }
        if ( k == NSTEPS )
            Hist[j][NSTEPS]++ ;
    }
}

// ***************************************
void CollectResults() 
{
    for ( int q = 1 ; q < 8 ; q++ )
#pragma GCC ivdep
        for ( int k = 0 ; k < NSTEPS ; k++ )
            Hist[0][k] += Hist[q][k] ; 
}



// ***************************************
void PrintResults() 
{
    unsigned long int max=0 ;
    int ind=0;

    for ( int k = 0 ; k < NSTEPS ; k++ )
        if ( max < Hist[0][k] ) {
            ind = k ;
            max = Hist[0][k] ;
        }
    printf( "Steps %d, Frequency %lu\n", ind, max) ;
}

// ***************************************
int main()
{
    pthread_t tid[8]; 
	int j ;

    for(  j = 0 ; j < 8 ; j++ ) {
    	pthread_create(&(tid[j]), NULL, CollatzSerial, (void *) (&j)); 
	}
    for(  j= 0 ; j < 8 ; j++ )  
    	pthread_join(tid[j], NULL);

    CollectResults() ;
    PrintResults() ;
    return 0 ;
}
